/**
 * @file   veil_query.c
 * \code
 *     Author:       Marc Munro
 *     Copyright (c) 2005, 2006 Marc Munro
 *     License:      BSD
 * $Id: veil_query.c,v 1.4 2006/11/01 00:08:55 bloodnok Exp $
 * \endcode
 * @brief  
 * Functions to simplify SPI-based queries.  These are way more
 * sophisticated than veil really needs but are nice and generic.
 * 
 */


#include <stdio.h>
#include "postgres.h"
#include "executor/spi.h"
#include "veil_funcs.h"
#include "veil_version.h"

/**
 * A Fetch_fn is a function that processes records, one at a time,
 * returned from a query.
 */
typedef bool (Fetch_fn)(HeapTuple, TupleDesc, void *);



/**
 * The number of records to fetch in one go from the query executor.
 */
#define FETCH_SIZE    20



/** 
 * Prepare a query for query().  This creates and opens a cursor.  The
 * caller must have established SPI_connect.  It is assumed that no
 * parameters to the query will be null.
 * \param qry The text of the SQL query to be performed.
 * \param nargs The number of input parameters ($1, $2, etc) to the query
 * \param argtypes Pointer to an array containing the OIDs of the data
 * \param args Actual parameters
 * types of the parameters 
 * \param read_only Whether the query should be read-only or not
 * \param saved_plan Adress of void pointer into which the query plan
 * will be saved.  Passing the same void pointer on a subsequent call
 * will cause the saved query plan to be re-used.
 * \return A portal containing the opened cursor.
 */
static Portal
prepare_query(const char *qry,
			  int nargs,
			  Oid *argtypes,
			  Datum *args,
			  bool read_only,
			  void **saved_plan)
{
    void   *plan;
    Portal  portal;
    int     prep_ok;
	
    if (saved_plan && *saved_plan) {
		// A previously prepared plan is available, so use it
		plan = *saved_plan;
    }
    else {
		if (!(plan = SPI_prepare(qry, nargs, argtypes))) {
			ereport(ERROR,
					(errcode(ERRCODE_INTERNAL_ERROR),
					 errmsg("prepare_query fails"),
					 errdetail("SPI_prepare('%s') returns NULL "
							   "(SPI_result = %d)", 
							   qry, SPI_result)));
		}

		if (saved_plan) {
			// We have somewhere to put the saved plan, so save it.
			*saved_plan = SPI_saveplan(plan);
		}
    }
	
    if ((portal = SPI_CURSOR_OPEN(NULL, plan, args, 
								  NULL, read_only)) == NULL) {
		ereport(ERROR,
				(errcode(ERRCODE_INTERNAL_ERROR),
				 errmsg("prepare_query fails"),
				 errdetail("SPI_cursor_open('%s') returns NULL "
						   "(SPI_result = %d)", 
						   qry, SPI_result)));
    }
    return portal;
}


/** 
 * Prepare and execute a query.  Query execution consists of a call to
 * process_row for each returned record.  Process_row can return a
 * single value to the caller of this function through the fn_param
 * parameter.  It is the caller's responsibility to establish an SPI
 * connection with SPI_connect.  It is assumed that no parameters to
 * the query, and no results will be null.
 * \param qry The text of the SQL query to be performed.
 * \param nargs The number of input parameters ($1, $2, etc) to the query
 * \param argtypes Pointer to an array containing the OIDs of the data
 * \param args Actual parameters
 * types of the parameters 
 * \param read_only Whether the query should be read-only or not
 * \param saved_plan Adress of void pointer into which the query plan
 * will be saved.  Passing the same void pointer on a subsequent call
 * will cause the saved query plan to be re-used.
 * \param process_row  The ::Fetch_fn function to be called for each
 * fetched row to process it.  If this is null, we simply count the row,
 * doing no processing on the tuples returned.
 * \param fn_param  An optional parameter to the process_row function.
 * This may be used to return a value to the caller.
 * \return The total number of records fetched and processed by
 * process_row.
 */
static int
query(const char *qry,
      int nargs,
      Oid *argtypes,
      Datum *args,
	  bool  read_only,
      void *saved_plan,
      Fetch_fn process_row,
      void *fn_param)
{
    int    row;
    int    total_rows = 0;
	int    fetched;
    bool   continuing = true;
    Portal portal = prepare_query(qry, nargs, argtypes, args, 
								  read_only, saved_plan);
	
    while (continuing) {
		SPI_cursor_fetch(portal, true, FETCH_SIZE);
		fetched = SPI_processed;

		for(row = 0; continuing; row++) {
			if (row >= fetched) {
				continuing = false;
			}
			else {
				// Process a row using the processor function
				if (process_row) {
					continuing = process_row(SPI_tuptable->vals[row], 
											 SPI_tuptable->tupdesc,
											 fn_param);
				}
				total_rows++;
			}
		}
    }
    SPI_cursor_close(portal);
    return total_rows;
}

/** 
 * ::Fetch_fn function for processing a single row of a single integer for 
 * ::query.
 * \param tuple The row to be processed
 * \param tupdesc Descriptor for the types of the fields in the tuple.
 * \param p_result Pointer to an int4 variable into which the value
 * returned from the query will be placed.
 * \return false.  This causes ::query to terminate after processing a
 * single row.
 */
static bool
fetch_one_int(HeapTuple tuple, TupleDesc tupdesc, void *p_result)
{
    int4 col = DatumGetInt32(SPI_getbinval(tuple, tupdesc, 1, false));
    *((int4 *) p_result) = col;
	
    return false;
}

/** 
 * ::Fetch_fn function for processing a single row of a single integer for 
 * ::query.
 * \param tuple The row to be processed
 * \param tupdesc Descriptor for the types of the fields in the tuple.
 * \param p_result Pointer to an int4 variable into which the value
 * returned from the query will be placed.
 * \return false.  This causes ::query to terminate after processing a
 * single row.
 */
static bool
fetch_one_bool(HeapTuple tuple, TupleDesc tupdesc, void *p_result)
{
    bool col = DatumGetBool(SPI_getbinval(tuple, tupdesc, 1, false));
    *((bool *) p_result) = col;
	
    return false;
}

/** 
 * ::Fetch_fn function for processing a single row of a single integer for 
 * ::query.
 * \param tuple The row to be processed
 * \param tupdesc Descriptor for the types of the fields in the tuple.
 * \param p_result Pointer to an int4 variable into which the value
 * returned from the query will be placed.
 * \return false.  This causes ::query to terminate after processing a
 * single row.
 */
static bool
fetch_one_str(HeapTuple tuple, TupleDesc tupdesc, void *p_result)
{
    char *col = SPI_getvalue(tuple, tupdesc, 1);
	char **p_str = (char **) p_result;
	*p_str = col;
	
    return false;
}

/** 
 * Executes a query that returns a single int4 value.
 * 
 * @param qry The text of the query to be performed.
 * @param result Variable into which the result of the query will be placed.
 * 
 * @return true if the query returned a record, false otherwise.
 */
static bool
int_from_query(const char *qry,
			   int4 *result)
{
	int     rows;
    Oid     argtypes[0];
    Datum   args[0];
	rows = query(qry, 0, argtypes, args, true, NULL, 
				 fetch_one_int, (void *)result);
	return (rows > 0);
}

/** 
 * Executes a query that returns a single bool value.
 * 
 * @param qry The text of the query to be performed.
 * @param result Variable into which the result of the query will be placed.
 * 
 * @return true if the query returned a record, false otherwise.
 */
bool
vl_bool_from_query(const char *qry,
				bool *result)
{
	int     rows;
    Oid     argtypes[0];
    Datum   args[0];
	rows = query(qry, 0, argtypes, args, true, NULL, 
				 fetch_one_bool, (void *)result);
	return (rows > 0);
}

/** 
 * Executes a query by oid, that returns a single string value.
 * 
 * @param qry The text of the query to be performed.
 * @param param The oid of the row to be fetched.
 * @param result Variable into which the result of the query will be placed.
 * 
 * @return true if the query returned a record, false otherwise.
 */
static bool
str_from_oid_query(const char *qry,
				   const Oid param,
				   char *result)
{
	int     rows;
    Oid     argtypes[1] = {OIDOID};
    Datum   args[1];
	
	args[0] = ObjectIdGetDatum(param);
	rows = query(qry, 1, argtypes, args, true, NULL, 
				 fetch_one_str, (void *)result);
	return (rows > 0);
}

/** 
 * Raise an error if type is not as expected.
 * 
 * @param oid Type oid of value.
 * @param expected_oid Type oid of value that we are expecting.
 * @param msg Supplementary text for error message if the types do not match.
 */
static void
check_type(Oid oid,
		   Oid expected_oid,
	       const char *msg)
{
	if (oid != expected_oid) {
		char *expected;
		char *actual;
		bool  ok;
		if (str_from_oid_query("select typname from pg_type where oid = $1",
							   expected_oid, (void *) &expected) &&
			str_from_oid_query("select typname from pg_type where oid = $1",
							   oid, (void *) &actual))
		{
			ereport(ERROR,
					(errcode(ERRCODE_INTERNAL_ERROR),
					 errmsg("Type mismatch"),
					 errdetail("Expected %s, got %s): %s", 
							   expected, actual, msg)));
		}
		else {
			// Failed to read db, so put out a simpler message
			ereport(ERROR,
					(errcode(ERRCODE_INTERNAL_ERROR),
					 errmsg("Type mismatch"),
					 errdetail("Incorrect type: %s", msg)));
		}
	}
}
		   
/** 
 * Determine whether the given oid represents an existing database or not.
 * 
 * @param dbid Oid of the database in which we are interested.
 * 
 * @result True if the database exists.
 */

extern bool
vl_db_exists(Oid db_id)
{
	char dbname[NAMEDATALEN + 1];

	return str_from_oid_query("select datname from pg_database where oid = $1",
							  db_id, dbname);
}


